var modulesPath = '/modules/';

function TMA() {

	// store this object to access it in xhr-callbacks (where "this" has another scope)
	var self = this;

	// default modules configuration
	this.defaultModules = 'menumanager,channelspage,timerspage,timeredit,recordingspage,dirselector,recfileinfo,epgpage,eventinfo,controlpage,logpage,configpage';

	// Debug Logs erzeugen
	this.debug = false;

	this.initStatus = {
			config: false,
			language: false,
			taps: false,
			settings: false,
			modules: false,
			logos: false,
	};

	/**
	 * contains the language json response
	 */
	this.i18 = null;

	/**
	 * contains the config json response
	 */
	this.config = null;

	/**
	 * contains the difference between PVR-Time and Client-Time
	 * (calculated in apiGetConfig)
	 */
	this.timeOffset = 0;

	/**
	 * contains the channels json response
	 */
	this.channelsObj = null;

	/**
	 * private associative array[channelid]
	 * access through method channel(channelid)
	 * @type {Object} channel object
	 */
	var channels = null;

	/**
	 * private associative array[transponderindex]
	 * access through method bouquet(transponderindex)
	 * @type {Object} bouquet object
	var bouquets = null;
	 */

	/**
	 * contains the favourites json response
	 */
	this.favourites = null;

	/**
	 * contains the files json response
	 */
	this.files = null;

	/**
	 * contains the timers json response
	 */
	this.timers = null;

	/**
	 * contains the logos json response
	 */
	this.logos = null;

	/**
	 * contains the settings json response
	 */
	var settings = null;

	/**
	 * curently running taps
	 */
	var taps = null;

	/**
	 * current WebControl version as number
	 */
	this.webControlVersion = 0;

	this.progressMax = 0;
	this.progressCnt = 0;

	this.directories = new Array();

	/**
	 * URL to get the screen image, default is bmp, but Safari needs tif ( see init() )
	 */
	this.PVRscreen = (
			// detect Safari browser
			// Safari only reports itself as Safari.
			// Safari needs .tif graphics instead of .bmp
			(navigator.userAgent.search("Safari") >= 0)	&&
			(navigator.userAgent.search("Chrome") < 0)			// Opera reports itself as both Safari AND Chrome
	) ?
			"/@@screen.tif"
			: '/@@screen.bmp';

	/**
	 * extended methods
	 */
	this.fn = new tmaModulManager(this);

	/**
	 * registers a module
	 *
	 * @param {tmaModule}	modulConstructor	constructor function
	 */
	this.registerModule = function(modulConstructor) {

		this.fn.registerModule(modulConstructor);

	};

	/**
	 * registers a manifest
	 *
	 * @param {String}	moduleName
	 * @param {Object}	manifest	JSON Object
	 */
	this.registerManifest = function(moduleName, manifest) {

		this.fn.registerManifest(moduleName, manifest);

	};

	/**
	 * checks, whether any module handles a specific feature
	 *
	 * @param {String}		featureName
	 * @returns {Boolean}
	 */
	this.featureExist = function( featureName ) {

		return this.fn.featureExist( featureName );

	};

	/**
	 * add an API method in tma
	 *
	 * @param {Object}	handlerObj
	 * @param {String}	methodName	must start with api
	 */
	this.addApiExtension = function(handlerObj, methodName) {

		if (
				(methodName.indexOf('api') === 0) &&
				(!this[methodName])
		) {

			// create a function, that can be directly called with tma.fn.<methodName>
			this[methodName] = function() {

				//var args = $.extend([], arguments);
				return handlerObj[methodName].apply(handlerObj, arguments /*args*/);

			};

		}

	};

	var log = new Array();

	/**
	 * log to the console and trigger a log event
	 *
	 * @param msg	string or object of type
	 * 	{
	 * 		title: log-group title
	 * 		details:{
	 * 			caption: value,
	 * 			...
	 * 		}
	 * 	}
	 */
	this.log = function (msg, modulename) {


		if (typeof console != "undefined") {

			if (typeof msg == 'string') {
				// log string
				console.log(msg);
			} else if (typeof msg == 'object') {
				// log group
				var prefix = '';
				if (console.group) {
					console.group(msg.title);
				} else {
					// console.group is not available
					console.log(msg.title);
					prefix = '- ';
				}
				$.each(msg.details, function(caption, value) {
					if (typeof value != 'undefined') {
						console.log(prefix + caption + ': ' + value);
					}
				});
				if (console.group) console.groupEnd();

			}

		}

		var s = '';
		if (typeof msg == 'string') {
			// log string
			s = msg;
		} else if (typeof msg == 'object') {
			// log group
			s = msg.title + '\n';
			$.each(msg.details, function(caption, value) {
				if (typeof value != 'undefined') {
					s += '  ' + caption + ': ' + value + '\n';
				}
			});
		}
		s += '\n';

		log.push(s);

		if (this.webControlVersion >= 53) {

			var url = '/api?function=log&action=write!';
			if (modulename) url += 'module=' + modulename;

			$.post(url, s);

		}

		$(this).triggerHandler('tmaLog', [s]);

	};

	/**
	 * get stored log messages
	 *
	 * @returns {Array}
	 */
	this.getLog = function() {

		return log;

	};

	/**
	 * triggers a tmaProgress event
	 */
	this.progress = function(msg) {

		this.progressCnt++;
		//this.log( 'progress: {0} / {1} - {2}'.format( this.progressCnt, this.progressMax, msg ) );
		$(this).triggerHandler('tmaProgress', [ this.progressCnt * 100 / this.progressMax, msg]);
		if ( this.progressMax == this.progressCnt ) {
			this.progressMax = 0;
			this.progressCnt = 0;
		}

	};

	/**
	 * Get the file "language-??.txt" from the PVR where "??" is the PVR's local language
	 *
	 * @param {String}		modulename	(optional) get the language file of a specific module
	 * @returns {Promise}	(allways resolved)
	 */
	this.apiGetLanguage = function (modulename) {

		var deferred = $.Deferred();

		var params = {
				'function':	'language',
				action:		'get!'
		};

		// language option
		if (this.localSetting('languagemode', 'pvr') != 'pvr') {
			params.lng = this.localSetting('languagemode', 'pvr');
		}

		// module option
		if (modulename) {
			params.module = modulename;
		} else {
			this.progressMax++;
		}

		$.getJSON('/api', params, function(lang) {

			if (!modulename) {

				// global lng file
				self.i18 = lang;
				self.initStatus.language = true;

			}

			deferred.resolve(lang);

		}).fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.apiGetLanguage()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
					parameter: '',
					modulename: modulename,
					language: params.lng,
				}
			});

			deferred.resolve({}); // return an empty language result

		}).always(function() {

			tma.progress( (modulename || 'tma') + '.language' );

		});

		return deferred.promise();

	};

	/**
	 * load TMAs settings list
	 *
	 * @returns jqXHR
	 */
	this.apiListSettings = function (modulename) {

		var params = {
				'function':	'setting',
				action:		'list!',
		};
		if (modulename) {
			params.module = modulename;
		} else {
			this.progressMax++;
		}

		return $.getJSON('/api', params).done(function(data) {

			if (!modulename) {
				settings = data.setting;
				self.initStatus.settings = true;
			}

		}).fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.apiListSettings()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
					parameter: '',
					modulename: modulename,
				}
			});

		}).always(function() {

			self.progress( (modulename || 'tma') + '.settings');

		});

	};

	/**
	 * load a WebControl config value
	 *
	 * @param key		name of the config param
	 * @param _default	default value, if setting doesnt exist
	 * @returns any value
	 */
	this.setting = function (key, _default) {

		if ( (!settings) || (settings[key] == undefined) ) {
			return _default;
		}

		return settings[key];

	};

	/**
	 * save a WebControl config value
	 *
	 * @param key		name of the config param
	 * @param value
	 * @returns {jqXHR}
	 */
	this.setSetting = function (key, value) {

		settings[key] = value;

		var params = {
				'function':	'setting',
				action:		'set!',
				key:		key,
				value:		value
		};

		return $.getJSON('/api', params).fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.setSetting()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
					parameter: '',
					key: key,
					value: value,
				}
			});

		});

	};

	/**
	 * load a local config value
	 *
	 * @param key		name of the config param
	 * @param _default	default value, if setting doesnt exist
	 * @returns any value
	 */
	this.localSetting = function (key, _default) {

		/*if ( $.jStorage.currentBackend() ) {
			return $.jStorage.get(key, _default);
		}*/

		return amplify.store( key ) || _default;

	};

	/**
	 * save a local config value
	 *
	 * @param key		name of the config param
	 * @param value
	 */
	this.setLocalSetting = function (key, value) {

		/*if ( $.jStorage.currentBackend() ) {
			$.jStorage.set(key, value);
		}*/

		amplify.store( key, value );

	};

	/**
	 * get modules configuration (local)
	 *
	 * @returns {Array}
	 */
	this.modulesToLoad = function() {

		return tma.localSetting('modules', this.defaultModules).split(',');

	};

	/**
	 * Get the config of the PVR
	 *
	 * @returns jqXHR
	 */
	this.apiGetConfig = function () {

		var params = {
				'function':	'config',
				action:		'get!'
		};

		this.progressMax++;

		return $.getJSON('/api', params).done(function(data) {

			self.config = data;
			//Calculate the time difference between the PVR and the browser.  This will be used for the default start time when creating new timers.
			var pvrNow = new Date(data.timestamp);
			var clientNow = new Date();
			self.timeOffset = clientNow - pvrNow;   //This is the number of milliseconds difference between the local time and the PVR time.

			//var testdate = new Date( new Date() - self.timeOffset);
			//alert("PVR current time = " + pvrNow + "\nLocal current time = " + clientNow + "\nOffset = " + self.timeOffset + "\nCalculated PVR time = " + testdate);

			if ( data.modules ) {
				$.each( data.modules, function(i, module) {
					self.fn.addModule(module.name, module.name_hex);
				});
			}

			// get the TMA1 version as a number
			self.webControlVersion = parseInt( self.config.tma1ver.replace(/\D+/g, '') );
			//console.log('webControlVersion: ' + self.webControlVersion);

			self.initStatus.config = true;

			self.progress('tma.config');

		}).fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.apiGetConfig()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
				}
			});

			self.progress('tma.config ' + error);

		});

	};

	/**
	 * Get a listing of all channels from the PVR
	 *
	 * @returns jqXHR
	 */
	this.apiGetChannels = function () {

		var params = {
				'function':	'channel',
				action:		'get!'
		};

		this.progressMax++;

		return $.getJSON('/api', params).done(function(data) {

			if (!channels) {

				self.channelsObj = data;

				channels = {};

				$.each(self.channelsObj.channels, function(nr, channel) {
					//channels[channel.chtype + 1].push(channel);
					channels[channel.logoid] = channel;
				});

			}

			self.progress('tma.channels');

		}).fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.apiGetChannels()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
				}
			});

			self.progress('tma.channels ' + error);

		});

	};

	/**
	 * return a listing of channels
	 *
	 * @param {String} favGroup		if empty, return all channels
	 * 								virtual favGroups can be __analogue, __tv, __radio
	 * @returns {Array} of channel objects
	 */
	this.getChannels = function (favGroup) {

		var chlist = new Array();

		if (!favGroup) {
			chlist = this.channelsObj.channels;
		} else if (favGroup == '__analogue') {
			// return all analogue channels
			$.each(this.channelsObj.channels, function(nr, channel) {
				if (channel.chtype == -1) {
					chlist.push(channel);
				}
			});
		} else if (favGroup == '__tv') {
			// return all tv channels
			$.each(this.channelsObj.channels, function(nr, channel) {
				if (channel.chtype == 0) {
					chlist.push(channel);
				}
			});
		} else if (favGroup == '__radio') {
			// return all radio channels
			$.each(this.channelsObj.channels, function(nr, channel) {
				if (channel.chtype == 1) {
					chlist.push(channel);
				}
			});
		} else {
			// return channels of a favourits group
			$.each(this.favourites.groups, function(grpNr, group) {

				if (group.name == favGroup) {

					$.each(group.services, function(svcNr, service) {

						if (channels[service.logoid]) {
							chlist.push(channels[service.logoid]);
						}

					});

				}

			});
		}

		return chlist;

	};

	/**
	 * get a channel object through svcType and svcNum
	 *
	 * @param {String}	channelid (logoid)
	 * @returns {Object} channel
	 */
	this.channel = function(channelid) {

		if ( channels[channelid] ) {
			return channels[channelid];
		} else {
			return null;
		}

	};

	/**
	 * Get a listing of favourite-groups from the PVR
	 *
	 * @returns jqXHR
	 */
	this.apiGetFavourites = function () {

		var params = {
				'function':	'favourites',
				action:		'get!'
		};

		this.progressMax++;

		return $.getJSON('/api', params).done(function(data) {

			self.favourites = data;

			self.progress('tma.favourites');

		}).fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.apiGetFavourites()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
				}
			});

			self.progress('tma.favourites ' + error);

		});

	};

	var logoSpriteW = 0;
	var logoSpriteH = 0;
	var $logoStyle = null;

	this.refreshLogoStyles = function() {

		// remove old styles
		if ($logoStyle) {
			$logoStyle.remove();
		}

		var logodisplaywidth = self.setting('logodisplaywidth', 54);

		var bgW		= parseInt(logoSpriteW	* logodisplaywidth / 54);
		var logoW	= parseInt(54	* logodisplaywidth / 54);
		var bgH		= logoSpriteH;
		var logoH	= 34;

		var bgWs	= parseInt(bgW	* 0.5);
		var logoWs	= parseInt(logoW * 0.5);
		var bgHs	= parseInt(bgH	* 0.5);
		var logoHs	= parseInt(logoH * 0.5);
		// create css-rules
		var style =
			'<style type="text/css">' +
				'.channellogo {' +
					'background-size: ' + (bgW) + 'px ' + (bgH) + 'px;' +
					'width: '  + (logoW) + 'px;' +
					'height: ' + (logoH) + 'px;' +
				'}\n' +
				'.channellogo-small {' +
					'background-size: ' + (bgWs) + 'px ' + (bgHs) + 'px;' +
					'width: '  + (logoWs) + 'px;' +
					'height: ' + (logoHs) + 'px;' +
				'}\n';

		// add logo classes
		$.each(self.logos.logos, function(nr, logo) {

			if (channels[logo.logoid]) {

				var x = - parseInt(logo.sprite_x * logodisplaywidth / 54);
				var y = - logo.sprite_y;
				style += '.logo-' + logo.logoid + ' { background-position: ' + x + 'px ' + y + 'px;}\n';

				x = parseInt(x * 0.5);
				y = parseInt(y * 0.5);
				style += '.logo-small-' + logo.logoid + ' { background-position: ' + x + 'px ' + y + 'px;}\n';

			}

		});

		style += '</style>';

		$logoStyle = $(style).appendTo('HEAD');

	};

	/**
	 * Get a listing of logos from the PVR
	 *
	 * @returns {Promise}	(allways resolved)
	 */
	this.apiGetLogos = function () {

		var deferred = $.Deferred();

		var params = {
				'function':	'logos',
				action:		'get!'
		};

		this.progressMax++;

		$.getJSON('/api', params).done(function(data) {

			self.logos = data;

			// assign logos to channels
			$.each(data.logos, function(nr, logo) {

				if (channels[logo.logoid]) {

					channels[logo.logoid].logo = logo;

				}

			});

			$('<img src="/service-logos.png">').one('load', function() {

				// the image is loaded
				logoSpriteW = this.naturalWidth  || this.width;
				logoSpriteH = this.naturalHeight || this.height;

				self.refreshLogoStyles();

			}).each(function() {
				if(this.complete) $(this).load();
			});

			self.initStatus.logos = true;
			self.progress('tma.logos');

		}).fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.apiGetLogos()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
				}

			});

			self.progress('tma.logos ' + error);

		}).always(function() {
	
			deferred.resolve();
			
		});

		return deferred.promise();

	};

	/**
	 * Select a service to view
	 *
	 * @param {Number} chType		0 = TV, 1 = Radio, -1 = Analogue
	 * @param {Number} chNum		channel number startin from 0
	 * @returns jqXHR
	 */
	this.apiChannelSet = function (chType, chNum) {

		/**
		 * this construct allows us to switch to fail(),
		 * if the api returns an error_code <br>
		 * see: return deferred.promise().fail()
		 * http://stackoverflow.com/a/5112734
		 */
		var deferred = $.Deferred();

		var params = {
				'function':	'channels',
				action:		'set!',
				chtype:		chType,
				chnum:		chNum
		};

		$.getJSON('/api', params).done(function(data, status, xhr) {

			if (data.status === 0) {
				// ok
				deferred.resolve(data, status, xhr);
			} else {
				// switch to fail()
				deferred.reject(
						xhr,
						'tma-error',
						data.statustext
				);
			}

		}).fail(function(xhr, errorType, error) {
			deferred.reject(xhr, errorType, error);
		});

		return deferred.promise().fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.apiChannelSet()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
					parameter: '',
					chType: chType,
					chNum: chNum,
				}
			});

		});

	};

	/**
	 * send a remote key
	 *
	 * @param {String} key
	 * @returns jqXHR
	 */
	this.sendremote = function (key) {

		var params = {
				'function':	'remote',
				action:		'send!',
				key:		key
		};

		return $.getJSON('/api', params).fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.sendremote()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
					parameter: '',
					key: key,
				}
			});

		});

	};

	/**
	 * return the sort function
	 * (private method)
	 *
	 * @param {String} key			property name of the file object
	 * @param {String} direction	'asc' or 'desc'
	 * @returns {Function}
	 */
	function getSortFilesFn(key, direction) {

		function getProperty(obj, prop) {

			var path = prop.split('.');

			if (path.length > 1) {
				var key = path.shift();
				var subobj = obj[key];
				if (typeof subobj == 'object') {
					return getProperty(subobj, path.join('.'));
				} else {
					return "";
				}

			} else {
				return obj[path[0]];
			}

		}

		var dir = (direction == 'desc') ? -1 : 1;

	    return function(a, b) {

			var aAttr = parseInt(a.attr);
			var bAttr = parseInt(b.attr);

			if (aAttr < bAttr) {
				return dir;
			} else if (aAttr > bAttr) {
				return -dir;
			} else {

				var aVal = getProperty(a, key);
				var bVal = getProperty(b, key);
				if (typeof aVal == 'string') aVal = aVal.toLowerCase();
				if (typeof bVal == 'string') bVal = bVal.toLowerCase();

				if (aVal < bVal) {
					return -dir;
				} else if (aVal > bVal) {
					return dir;
				} else {
					return 0;
				}
			}

	    };
	}

	/**
	 * sort the filelist
	 *
	 * @param {String} key		property name of the file object
	 * @param {String} dir		sort direction: 'asc' or 'desc'
	 */
	this.sortFiles = function(key, dir) {

		this.files.files.sort( getSortFilesFn( key, dir ) );

	};

	/**
	 * Get a file listing from the PVR
	 *
	 * @param directory
	 * @returns jqXHR
	 */
	this.apiGetFiles = function(directory) {

		var params = {
				'function':	'files',
				action:		'get!',
				dir:		directory
		};

		return $.getJSON('/api', params, function(data) {

			self.files = data;

			// generete a value to sort for date

			for(var i = 0; i < self.files.files.length; i++) {

				var f = self.files.files[i];

				if ( (f.attr == 208) && (f.infpresent) && (f.infdata.header.starttime)) {
					f.sortdate = f.infdata.header.starttime;
				} else {
					f.sortdate = f.timestamp ;
				}
			}

		}).fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.apiGetFiles()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
					parameter: '',
					directory: directory,
				}
			});

		});

	};

	/**
	 * Get directory tree from the PVR
	 *
	 * @param {String} dir	tree root
	 * @returns jqXHR
	 */
	this.apiTreeFiles = function(dir) {

		var params = {
				'function':	'files',
				action:		'tree!'
		};
		if (dir) params.dir = dir;

		return $.getJSON('/api', params, function(data) {

			self.directories = new Array();

			$.each(data.directory, function(i, directory) {

				var dir = new TmaDir(directory.name);
				dir.insert(self.directories);

			});

		}).fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.apiTreeFiles()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
					parameter: '',
					directory: dir,
				}
			});

		});

	};

	/**
	 * Get the status object from the PVR
	 * @returns jqXHR
	 */
	this.apiGetStatus = function() {

		var params = {
				'function':	'status',
				action:		'get!'
		};

		return $.getJSON('/api', params).fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.apiGetStatus()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
				}
			});

		});

	};

	/**
	 * Get the running taps from the PVR
	 *
	 * @returns jqXHR
	 */
	this.apiGetTaps = function() {

		var deferred = $.Deferred();

		var params = {
				'function':	'taps',
				action:		'get!'
		};

		this.progressMax++;

		deferred.then(function() {

			$.getJSON('/api', params).done(function(tapObject) {

				taps = tapObject.tap;

				self.initStatus.taps = true;

				self.progress('tma.taps');

			}).fail(function(xhr, errorType, error) {

				self.log({
					title: errorType + " in tma.apiGetTaps()",
					details: {
						error: error,
						xhrReadyState: xhr.readyState,
						xhrStatus: xhr.status,
					}
				});

				self.progress('tma.taps ' + error);

			});

		});

		deferred.resolve();

		return deferred.promise();

	};

	/**
	 * tests, if a tap is running
	 */
	this.isTapRunning = function(tapid) {

		for( var i = 0; i < taps.length; i++) {

			if (taps[i].tapid == tapid)	return true;
		}

		return false;

	};

	/**
	 * plays a recording
	 *
	 * @param {String} directoryToPlay	path
	 * @param {String} fileToPlay		filename
	 * @param {String} hexToPlay		hex representation of the filename
	 * @param {Number} startBlock		blocksize = config.blockfactor  *** DMC - This is wrong, the block factor is something else.  Use "startposition"
	 * @returns jqXHR
	 */
	this.apiPlayFile = function (directoryToPlay, fileToPlay, hexToPlay, startBlock) {

		var params =
				"dir="	    + directoryToPlay + "\n" +
				"file="	    + fileToPlay + "\n" +
				"block="    + startBlock + "\n" +
				"file_hex=" + hexToPlay;

		return $.post('/api?function=files&action=play!', params, null, 'json').fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.apiPlayFile()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
					parameter: '',
					directoryToPlay: directoryToPlay,
					fileToPlay: fileToPlay,
					hexToPlay: hexToPlay,
					startBlock: startBlock,
				}
			});

		});

	};

	/**
	 * stops a playback
	 *
	 * @returns jqXHR
	 */
	this.stopPlayback = function () {

		var params = {
				'function':	'files',
				action:		'stop!'
		};

		return $.getJSON('/api', params).fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.stopPlayback()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
				}
			});

		});

	};

	/**
	 * stops a recording
	 *
	 * @param {Number} slot
	 * @returns jqXHR
	 */
	this.stopRecording = function (slot) {

		var params = {
				'function':	'files',
				action:		'stoprec!',
				slot: 		slot
		};

		return $.getJSON('/api', params).fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.stopRecording()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
					parameter: '',
					slot: slot,
				}
			});

		});

	};

	/**
	 * delete a recording
	 *
	 * @param {String} directory
	 * @param {String} filename
	 * @param {String} filename_hex		hex representation of the filename
	 * @returns jqXHR
	 */
	this.apiDeleteFile = function (directory, filename, filename_hex) {

		var params =
			"file="	+ filename + "\n" +
			"dir="	+ directory + "\n" +
			"file_hex="	+ filename_hex + "\n";

		return $.post('/api?function=files&action=delete!', params, null, 'json').done(function(delObject) {

			if (delObject.mpg_status == 0) {

				// deleted => remove from files
				for ( var i = 0; i < self.files.files.length; i++) {

					var f = self.files.files[i];
					if (f.name_hex == filename_hex /* delObject.file_hex */) {
						self.files.count--;
						self.files.files.splice(i, 1);
						break;
					}

				}

			}

		}).fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.apiDeleteFile()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
					parameter: '',
					directory: directory,
					filename: filename,
					filename_hex: filename_hex,
				}
			});

		});

	};

	/**
	 * move/rename a recording
	 *
	 * @param {String} directory
	 * @param {String} filename
	 * @param {String} filename_hex		hex representation of the filename
	 * @param {String} newDirectory
	 * @param {String} newFilename
	 * @returns jqXHR
	 */
	this.apiRenameFile = function (directory, filename, filename_hex, newDirectory, newFilename) {

		var params =
			"dir="	+ directory + "\n" +
			"file="	+ filename + "\n" +
			"file_hex="	+ filename_hex + "\n" +
			"new_dir="	+ newDirectory + "\n" +
			"new_file="	+ newFilename + "\n";

		return $.post('/api?function=files&action=rename!', params, null, 'json').done(function(moveObject) {

			if (
					(moveObject.mpg_status == 0) &&
					(moveObject.inf_status == 0) &&
					(moveObject.nav_status == 0)
			) {

				// keep files up to date
				for ( var i = 0; i < self.files.files.length; i++) {

					var f = self.files.files[i];
					if (f.name_hex == filename_hex /*renameObject.old_file*/) {
						if (directory == newDirectory) {
							// same directory => rename
							f.name = moveObject.new_file;
							// @todo refresh name_hex
							f.name_hex = moveObject.new_file_hex;
						} else {
							// different directory => remove
							self.files.count--;
							self.files.files.splice(i, 1);
						}
						break;
					}

				}

			} else {

				self.log("ERROR in tma.apiRenameFile(" + directory + ", " + filename + ", " + newDirectory + ", " + newFilename + "):");
				self.log('MPG: status: ' + moveObject.mpg_status + '; errno: ' + moveObject.mpg_errno);
				self.log('INF: status: ' + moveObject.inf_status + '; errno: ' + moveObject.inf_errno);
				self.log('NAV: status: ' + moveObject.nav_status + '; errno: ' + moveObject.nav_errno);

			}

		}).fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.moveFile()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
					parameter: '',
					directory: directory,
					filename: filename,
					filename_hex: filename_hex,
					newDirectory: newDirectory,
					newFilename: newFilename,
				}
			});

		});

	};

	//Used for the sort routine when timers are returned
	function timerCompare(a, b)  {

		if (a.starttime < b.starttime) return -1;
		if (a.starttime > b.starttime) return 1;
		return 0;

	}

	/**
	 * Get a timers listing from the PVR
	 *
	 * @returns jqXHR
	 */
	this.apiGetTimers = function() {

		var params = {
				'function':	'timer',
				action:		'get!',
		};

		return $.getJSON('/api', params, function(timers) {

			self.timers = timers;
			
			// convert to TmaTimer objects
			for(var i = 0; i < self.timers.timers.length; i++) {
				
				self.timers.timers[i] = new TmaTimer(self.timers.timers[i]);
				
			} 

			self.timers.timers.sort(timerCompare);

		}).fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.apiGetTimers()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
				}
			});

		});

	};

	/**
	 * delete a timer
	 *
	 * @param {TmaTimer} _timer
	 * @returns jqXHR
	 */
	this.apiDeleteTimer = function (_timer) {

		var params =
			"index="			+ _timer.index			+ "\n" +
			"old_servicetype="	+ _timer.servicetype	+ "\n" +
			"old_serviceindex="	+ _timer.serviceindex	+ "\n" +
			"old_duration="		+ _timer.duration		+ "\n" +
			"old_starttime="	+ _timer.starttime		+ "\n";

		return $.post('/api?function=timer&action=delete!', params, null, 'json').done(function(delObject) {

			if (delObject.return_code == 0) {

				// find index:
				for (var i = 0; i < self.timers.timers.length; i++) {
					var timer = self.timers.timers[i];
					if (timer.index == _timer.index) {

						// remove from timers
						self.timers.count--;
						self.timers.timers.splice(i, 1);
						break;

					}
				}

			}

		}).fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.apiDeleteTimer()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
					parameter: '',
					timerIndex: _timer.index ,
					timerFilename: _timer.filename,
					timerServicetype: _timer.servicetype,
					timerServiceindex: _timer.serviceindex,
					timerDuration: _timer.duration,
					timerStarttime: _timer.starttime,
				}
			});

		});

	};

	/**
	 * update/create a timer
	 *
	 * @param {TmaTimer}	origTimer	Original timer (to provide verification parameters for the API).  Null for new timers.
	 * @param {TmaTimer}	newTimer	New timer details
	 * @returns jqXHR
	 */
	this.apiTimerUpdate = function (origTimer, newTimer) {

		var cmd = '/api?function=timer&action=create!';
		var params2 = '';

		if (origTimer.index >= 0) {
			cmd = '/api?function=timer&action=update!';
			params2 =
				"index="				+ origTimer.index			+ "\n" +
				"old_servicetype="		+ origTimer.servicetype		+ "\n" +
				"old_serviceindex="		+ origTimer.serviceindex	+ "\n" +
				"old_duration="			+ origTimer.duration		+ "\n" +
				"old_starttime="		+ origTimer.starttime		+ "\n";
		} else {
			if (newTimer.eventid1) {
	            //"eventid1"		- For a modification, the TAP does not change these values unless provided by the API
				params2 += "eventid1="	+ newTimer.eventid1			+ "\n";
			}
			if (newTimer.eventid2) {
				//"eventid2"		- For a new timer, these values can be provided if the timer was created from an EPG entry.
				params2 += "eventid2="	+ newTimer.eventid2			+ "\n";
			}
		}

		var params =
			"isrec="					+ newTimer.isrec			+ "\n" +
			"filename="					+ newTimer.filename			+ "\n" +
			"servicetype="				+ newTimer.servicetype		+ "\n" +
			"serviceindex="				+ newTimer.serviceindex		+ "\n" +
			"reservationtype="			+ newTimer.reservationtype	+ "\n" +
			"starttime="				+ newTimer.starttime		+ "\n" +
			"duration="					+ newTimer.duration			+ "\n" +
			params2 + "\n";

			//TODO - EventIDs
            //"epgmarker"		- This is calculated by the TAP, but can be overridden by the API is required.  1 = Play, 2 = Rec, 6 = IR.
            //"analoguetype"	- For PVRs that support analogue capture, 2 = Composite and 4 = Component

		/**
		 * this construct allows us to switch to fail(),
		 * if the api returns an error_code <br>
		 * see: return deferred.promise().fail()
		 * http://stackoverflow.com/a/5112734
		 */
		var deferred = $.Deferred();

		$.post(cmd, params, null, 'json').done(function(data, status, xhr) {

			/**
			 * return codes:
			 *  0 = Success,
			 *  1 = Create failed,
			 *  2 = key mismatch modify/delete,
			 *  3 = Index not set for modify/delete,
			 *  4 = Failed modify/delete
			 */

			if (data.return_code === 0) {

				// ok
				$.extend(origTimer, newTimer);

				if (origTimer.index < 0) {

					// really created new timer?
					if (data.start_count == data.end_count) {

						// switch to fail()
						deferred.reject(
								xhr,
								'tma-error',
								"Error creating timer (start_count=end_count). return_code:" + data.return_code
						);

					} else {

						//this is a new Timer
						origTimer.index = self.timers.timers.length;
						self.timers.timers.push(origTimer);
						self.timers.count++;
						$(self).triggerHandler('tmaTimerAdded', [origTimer]);

					}

				} else {

					$(self).triggerHandler('tmaTimerModified', [origTimer]);

				}

				deferred.resolve(data, status, xhr);

			} else {

				// switch to fail()
				deferred.reject(
						xhr,
						'tma-error',
						"Error creating/modifying timer. (" + data.return_code + " - " + data.return_text + ")"
				);

			}

		}).fail(function(xhr, errorType, error) {
			deferred.reject(xhr, errorType, error);
		});

		return deferred.promise().fail(function(xhr, errorType, error) {

			var details = {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
					parameterNewTimer: '',
					newIsrec:  newTimer.isrec,
					newFilename: newTimer.filename,
					newServicetype: newTimer.servicetype,
					newServiceindex: newTimer.serviceindex,
					newReservationtype: newTimer.reservationtype,
					newStarttime: newTimer.starttime,
					newDuration: newTimer.duration,
			};
			if (origTimer.index >= 0) {
				$.extend(details, {
					index: origTimer.index,
					parameterOldTimer: '',
					oldIsrec:  newTimer.isrec,
					oldFilename: newTimer.filename,
					oldServicetype: newTimer.servicetype,
					oldServiceindex: newTimer.serviceindex,
					oldReservationtype: newTimer.reservationtype,
					oldStarttime: newTimer.starttime,
					oldDuration: newTimer.duration,
				});
			}

			self.log({
				title: errorType + " in tma.apiTimerUpdate()",
				details: details,
			});

		});

	};

	/**
	 * load EPG data of a single channel
	 *
	 * @param {Number}		chtype			0: TV, 1: Radio
	 * @param {Number}		chnum			channel number
	 * @param {String}		starttime		in PVR date format
	 * @param {String}		endtime			in PVR date format
	 * @returns jqXHR
	 */
	this.apiGetEpg = function(chtype, chnum, starttime, endtime) {

		var params = {
				'function':	'epg',
				action:		'get!',
				chtype:		chtype,
				chnum:		chnum,
				starttime:	starttime,
				endtime:	endtime
		};

		return $.getJSON('/api', params).done( function(epg) {


		}).fail(function(xhr, errorType, error) {

			self.log({
				title: errorType + " in tma.apiGetEpg()",
				details: {
					error: error,
					xhrReadyState: xhr.readyState,
					xhrStatus: xhr.status,
					parameter: '',
					chtype: chtype ,
					chnum: chnum,
					starttime: starttime,
					endtime: endtime,
				}
			});

		});

	};

	/**
	 * load EPG data of multiple channels
	 *
	 * @param {Array}		chlist			Array of channel objects
	 * @param {Date}		starttime
	 * @param {Date}		endtime
	 * @param {Function}	channel_loaded	(optional) A callback function (channel, epgdata),
	 * 										which is called afer each channel EPG data is loaded
	 * @returns jQuery.Deferred	the last Deferred object of the query chain
	 */
	this.loadEpg = function(chlist, starttime, endtime, channel_loaded) {

		starttime	= this.toPvrDate(starttime);
		endtime		= this.toPvrDate(endtime);

		var deferred = $.Deferred();
		var pipe = deferred;

		$.each(chlist, function(ix, channel) {

			pipe = pipe.then(function() {
				return  self.apiGetEpg(channel.chtype, channel.chnum, starttime, endtime);
			}).done(function(epg) {
				//log(channel.servicename + ': ' + epg.count);
				if (channel_loaded) channel_loaded(channel, epg);
			});

			// test: reduce number of channels
			//if (ix == 20) return false;

		});

		deferred.resolve();

		return pipe;

	};

	/**
	 * get a two digit representation of a number
	 * @returns {String}
	 */
	function zeroPad(number) {
		return ( "0" + number ).substr( -2, 2 );
	};

	/**
	 * formats a date (partially) like PHP date() function <br>
	 * j:	Day of the month, without leading zeros 				1 to 31 <br>
	 * d:	Day of the month, 2 digits with leading zeros 			01 to 31 <br>
	 * D:	A textual representation of a day, three letters 		Sun through Mon <br>
	 * l:	A full textual representation of the day of the week 	Sunday through Saturday <br>
	 * m:	Numeric representation of a month, with leading zeros 	01 through 12 <br>
	 * M:	A short textual representation of a month, 3 letters 	Jan through Dec <br>
	 * F:	A full textual representation of a month			 	January through December <br>
	 * w:	Numeric representation of the day of the week		 	0 (for Sunday) through 6 (for Saturday) <br>
	 * Y:	A full numeric representation of a year, 4 digits 		Examples: 1999 or 2003 <br>
	 * y:	A two digit representation of a year 					Examples: 99 or 03 <br>
	 * g:	12-hour format of an hour without leading zeros 		1 through 12 <br>
	 * G:	24-hour format of an hour without leading zeros 		0 through 23 <br>
	 * h:	12-hour format of an hour with leading zeros 			01 through 12 <br>
	 * H:	24-hour format of an hour with leading zeros 			00 through 23 <br>
	 * i: 	Minutes with leading zeros 								00 to 59 <br>
	 * s: 	Seconds, with leading zeros 							00 through 59 <br>
	 *
	 * @param {String}	format	format-string
	 * @param {Date}	date
	 * @returns {String}
	 * @see {@link http://de3.php.net/manual/en/function.date.php}
	 */
	this.date = function(format, date) {

		if (typeof date != 'object') {
			date = new Date(date);
		}

		// http://de3.php.net/manual/en/function.date.php
		var dateMarkers = {
				j:['getDate'	, function(v) { return v;							}],
				d:['getDate'	, function(v) { return zeroPad(v);					}],
				D:['getDay'		, function(v) { return self.i18.days[v].shortname;	}],
				l:['getDay'		, function(v) { return self.i18.days[v].name;		}],
				m:['getMonth'	, function(v) { return zeroPad(v+1);				}],
				M:['getMonth'	, function(v) { return self.i18.months[v].shortname;}],
				F:['getMonth'	, function(v) { return self.i18.months[v].name;		}],
				w:['getDay'],
				y:['getFullYear', function(v) { return zeroPad(v);					}],
				Y:['getFullYear'],
				g:['getHours'	, function(v) { return ((v - 1) % 12) + 1;			}],
				G:['getHours'],
				h:['getHours'	, function(v) { return zeroPad( ((v - 1) % 12) + 1);}],
				H:['getHours'	, function(v) { return zeroPad(v);					}],
				i:['getMinutes'	, function(v) { return zeroPad(v);					}],
				s:['getSeconds'	, function(v) { return zeroPad(v);					}],
		};

		var dateTxt = format.replace(/[jdDlmMFwyYgGhHis]/g, function(m) {

			var rv = date[(dateMarkers[m])[0]]();

			if ( dateMarkers[m][1] != null ) rv = dateMarkers[m][1](rv);

			return rv;

		});

		return dateTxt;

	};

	/**
	 * format a duration to h:mm
	 *
	 * @param {Number}	theDuration
	 * @returns {String}
	 */
	this.formatDuration = function (theDuration) {
		dur = parseInt(theDuration / 60) + ":" + zeroPad(theDuration % 60);
		return dur;
	};

	/**
	 * get a PVR compatible date string ('yyyy/mm/dd hh:mm:ss')
	 *
	 * @param {Date}	date
	 * @returns {String}
	 */
	this.toPvrDate = function(date) {

		return this.date('Y/m/d H:i:s', date);

	};

	/**
	 * seperates the filename from a path
	 *
	 * @param {String}	path
	 * @returns {Object} { directory: , name: }
	 */
	this.splitPath = function(path) {

		var match = path.match( /^(\/mnt\/.*)\/([^\/]+)$/);
		if (match) {
			return {
				directory: match[1],
				name: match[2]
			};
		} else {
			return false;
		}

	};

	this.loadModules = function() {

		return self.fn.loadModules().done(function() {

			self.initStatus.modules = true;

		});

	};

	this.initChain = [
	                  this.apiGetConfig,
	                  this.apiGetLanguage,
	                  this.apiGetTaps,
	                  this.apiListSettings,
	                  this.fn.loadManifests,
	                  this.loadModules,
	                  this.apiGetChannels,
	                  this.apiGetFavourites,
	                  this.apiGetLogos
	                  ];

	/**
	 * load all data from the pvr, which never change during the session
	 *
	 * @returns {Promise} jQuery Promise Object http://api.jquery.com/Types/#Promise
	 */
	this.init = function() {

		var deferred = $.Deferred();
		var pipe = deferred;

		$.each(this.initChain, function(i, initFn) {

			pipe = pipe.then(function() {

				return initFn.call(self);

			});

		});

		deferred.resolve();

		pipe.promise().done(function() {

			// periodically check, wether all required modules were created
			var timer = null;
			timer = window.setInterval(function() {

				var created = self.fn.modulesCreated();
				if (self.debug) self.log('all modules created? ' + (created ? 'true' : 'false'));
				if (created) {

					window.clearTimeout(timer);

					self.fn.setupModules().done(function() {

						// init was successful
						$(self).triggerHandler('tmaReady');

					});

				}
			}, 100);

		});

	};

	/**
	 * decodes a hex representation of a filename
	 *
	 * @param {String}	name_hex
	 * @returns {String}
	 */
	this.decodeHexname = function(name_hex) {
		var result = '';

		while (name_hex.length) {

			var c = parseInt( '0x' + name_hex.substr(0,2) );

			if (c == 173) c = 45;	// '-'
			result += String.fromCharCode(c);

			name_hex = name_hex.substr(2);

		}

		return result;

	};

	/**
	 * removes the file extension (.mpg or .rec) from a filename
	 *
	 * @param {String}	filename
	 * @returns {String}
	 */
	this.removeExtension = function(filename) {

		return filename.replace( new RegExp( '\\' + self.config.recextension + '$' ), '' );

	};

}

var tma = new TMA();
