/**
 * class to manage moduls
 *
 * @param {TMA} tma
 * @returns {tmaModulManager}
 */
function tmaModulManager(tma) {

	/**
	 * moduls list by name
	 * objects of type tmaModuleData
	 */
	var moduleList = {};

	var self = this;

	/**
	 * list of tmaFeatures
	 */
	var tmaFeatures = {};

	/**
	 * list of lazy loaded files (URL array)
	 */
	var loadedFiles = [];

	/**
	 * add module data to the module list
	 * called only from tma.apiGetConfig for each available module
	 */
	this.addModule = function(name, nameHex) {

		if (!(moduleList[name])) {
			moduleList[name] = new tmaModuleData();
			moduleList[name].name = name;
		}
		moduleList[name].nameHex = nameHex;

	};

	/**
	 * returns a specific modul data
	 *
	 * @param {String}	name	module name
	 * @returns {tmaModuleData}	or null, if not found
	 */
	this.getModuleData = function(name) {

		if ( moduleList[name] ) {
			return moduleList[name];
		} else {
			return null;
		}

	};

	/**
	 * returns an array of available modules names
	 *
	 * @returns {Array}
	 */
	this.getModules = function() {

		var result = [];

		$.each(moduleList, function(name, moduleData) {
			result.push(name);
		});

		return result;


	};

	/**
	 * load all manifests of the required modules list
	 *
	 * @returns {Promise}	(allways resolved)
	 */
	this.loadManifests = function( ) {

		var deferred = $.Deferred();
		var pipe = deferred;

		var modulesToLoad = tma.modulesToLoad();

		$.each(modulesToLoad, function(i, name) {

			var moduleData = self.getModuleData(name);
			if (moduleData) {

				// load the manifest

				tma.progressMax++;

				pipe = pipe.then(function() {

					var pipeDeferred = $.Deferred();

					moduleData.loadManifest().always(function() {

						tma.progress(name + '.manifest');
						pipeDeferred.resolve();

					});

					return pipeDeferred.promise();

				});

			}

		});

		deferred.resolve();

		return pipe.promise();

	};

	/**
	 * load modules
	 *
	 * @returns {Promise} the last Promise object of the query chain
	 */
	this.loadModules = function() {

		var deferred = $.Deferred();
		var pipe = deferred;

		var modulesToLoad = tma.modulesToLoad();

		$.each(modulesToLoad, function( i, name ) {

			var moduleData = self.getModuleData(name);
			if (moduleData) {

				if (moduleData.isCreated()) {

					// static bound module

					// add manifest
					moduleData.module.manifest = moduleData.manifest;

				} else {

					// add to dynamic load pipe

					if ( moduleData.active ) {

						pipe = pipe.then(function() {

							return self.loadJs( moduleData.path() + name + '.js' );

						});

						tma.progressMax++;

					}

				}

			}

		});

		deferred.resolve();

		return pipe.promise();

	};

	/**
	 * check modules creation
	 *
	 * @returns {Boolean} are all active modules created?
	 */
	this.modulesCreated = function() {

		var result = true;

		var modulesToLoad = tma.modulesToLoad();

		$.each(modulesToLoad, function( i, name ) {

			var moduleData = self.getModuleData(name);
			if (moduleData) {

				if ( !moduleData.isCreated() && moduleData.active ) {

					result = false;
					return false; // break;
				}

			}

		});

		return result;

	};

	/**
	 * init all modules
	 *
	 * @returns {Promise}	(allways resolved)
	 */
	this.setupModules = function() {

		var deferred = $.Deferred();
		var pipe = deferred;

		var modulesToLoad = tma.modulesToLoad();

		$.each(modulesToLoad, function( i, name ) {

			var moduleData = self.getModuleData(name);
			if (moduleData) {

				if ( moduleData.isCreated() ) {

					// setup modul
					moduleData.module.path = moduleData.path();
					var setup = moduleData.module.setup( );

					if ( setup !== false ) {
						// setup returned a deferred object
						pipe = pipe.then(function() {
							return setup.done(function() {

								moduleData.setInitialised();

							});

						});

					}

				}

			}

		});

		deferred.resolve();

		return pipe.promise();

	};

	/**
	 * registers a manifest
	 *
	 * @param {String}	moduleName
	 * @param {Object}	manifest	JSON Object
	 */
	this.registerManifest = function(moduleName, manifest) {

		// create entry in moduleList
		if ( !this.getModuleData(moduleName) ) this.addModule(moduleName);

		var moduleData = self.getModuleData(moduleName);

		moduleData.manifest = manifest;		

	};

	/**
	 * registers a module
	 *
	 * @param {Function}	moduleConstructor
	 * @returns {tmaModule}	if registred
	 */
	this.registerModule = function(modulConstructor) {

		modulConstructor.prototype = new tmaModule();
		var module = new modulConstructor();

		// create entry in moduleList
		if ( !this.getModuleData(module.name) ) this.addModule(module.name);

		var moduleData = self.getModuleData(module.name);

		moduleData.setCreated();
		moduleData.module = module;

		if (moduleData.manifest) {

			// add manifest
			module.manifest = moduleData.manifest;

		}

		$(moduleData).triggerHandler('tmaModuleCreated', [module]);

		return module;

	};

	/**
	 * register an modul method to modul manager as a tmaFeature handler
	 *
	 * @param {Object}	handlerObj
	 * @param {String}	methodName
	 * @param {Boolean}	createMember	(optional default:true) create tma.fn.<methodName>
	 */
	this.addFeature = function(handlerObj, methodName, createMember) {

		if (!tmaFeatures[methodName]) {
			tmaFeatures[methodName] = new tmaFeature(methodName);

			if (createMember !== false) {
				if (tma.debug) tma.log('moduleManager.addFeature: ' + methodName);
				// create a function, that can be directly called with tma.fn.<methodName>
				this[methodName] = function() {

					if (tma.debug) tma.log('moduleManager.feature.' + methodName);
					return tmaFeatures[methodName].start( arguments );

				};
			}

		}

		tmaFeatures[methodName].addFeature(handlerObj);

	};

	/**
	 * checks, whether any module handles a specific tmaFeature
	 *
	 * @param {String}		featureName
	 * @returns {Boolean}
	 */
	this.featureExist = function( featureName ) {

		return true && (tmaFeatures[featureName]);

	};

	/**
	 * calls a specific tmaFeature
	 *
	 * @param {String}	featureName
	 * @param {Array}	args		optional parameters for the tmaFeature handlers
	 */
	this.runFeature = function( featureName, args ) {

		if (tmaFeatures[featureName]) {
			return tmaFeatures[featureName].start(args);
		}

	};

	/**
	 * load a JavaScript file, if not previosly loaded
	 *
	 * @returns {Promise}	(allways resolved)
	 */
	this.loadJs = function(url) {

		var deferred = $.Deferred();

		// remove "foldername/../", to get absolute path
		url = url.replace(/[^\/]+\/\.\.\//, '');

		if ( (url) && ($.inArray( url, loadedFiles ) < 0) ) {

			loadedFiles.push(url);

			$.ajax({
				dataType: "script",
				url: url,
			}).fail(function(xhr, errorType, error) {

				tma.log({
					title: errorType + " in tma.loadJs()",
					details: {
						error: error,
						xhrReadyState: xhr.readyState,
						xhrStatus: xhr.status,
						parameter: '',
						url: url,
					}
				});

			}).always(function() {
				tma.progress(url);
				deferred.resolve();
			});

		} else {
			deferred.resolve();
		}

		return deferred.promise();

	};

	/**
	 * load a css file, if not previosly loaded
	 *
	 * @param {String}	url
	 * @returns {Promise}	(allways resolved)
	 */
	this.loadCss = function (url) {

		var deferred = $.Deferred();

		// remove "foldername/../", to get absolute path
		url = url.replace(/[^\/]+\/\.\.\//, '');

		if ( (url) && ($.inArray( url, loadedFiles ) < 0) ) {

			loadedFiles.push(url);

			$.ajax({url : url, dataType: 'text', cache: false }).done(function(text) {

				$('<style type="text/css">' + text + '</style>').appendTo('head');

			}).fail(function(xhr, errorType, error) {

				tma.log({
					title: errorType + " in tma.loadCss()",
					details: {
						error: error,
						xhrReadyState: xhr.readyState,
						xhrStatus: xhr.status,
						parameter: '',
						url: url,
					}
				});

			}).always(function() {

				tma.progress(url);
				deferred.resolve();

			});

		} else {
			deferred.resolve();
		}

		return deferred.promise();

	};

}
