/**
 * menus managing module
 */
tma.registerModule( menumanager );
menumanager.prototype = new tmaModule();
function menumanager() {

	var self = this;
	
	// list of menu instances
	var menus = {};

	/**
	 * IMPORTANT: the modules name; equal to directory and script file name
	 */
	this.name = 'menumanager';

	/**
	 * initialize module
	 */
	this.setup = function() {

		// register tmaFeatures
		tma.fn.addFeature( this, 'addMenu' );
		tma.fn.addFeature( this, 'getMenu' );

		return this.parent().setup().done(function() {
			self.init();
		});

	};

	/**
	 * create menu instance
	 *
	 * @param {tmaFeature}	f
	 * @param {String}		id
	 */
	this.addMenu = function( f, id ) {

		f.stop();

		menus[id] = new MenuManager(id);

		/**
		 * let modules insert their register tabs
		 */
		tma.fn.runFeature('appendMenuTab', [ menus[id] ]);

	};

	/**
	 * return a menu instance for a given id
	 * if its not existing, a new instance will be created
	 *
	 * @param {tmaFeature}	f
	 * @param {String}		id
	 */
	this.getMenu = function( f, id ) {

		f.stop();
		
		if (!menus[id]) {
			menus[id] = new MenuManager(id);
		}

		return menus[id];
		
	};

}

/**
 * menu instance
 *
 * @param {String}	menuId
 * @returns {MenuManager}
 */
function MenuManager(menuId) {

	var self = this;

	var currentModule = null;

	this.menuId = menuId;

	/**
	 * adds a menu item for a page
	 * called from  modules, to add their pages
	 *
	 * @param {tmaModule}	module	the module should own these methods:
	 * 									show: callback to render the module content
	 * 									cleanup: callback to cleanup, what the module has modified outside content
	 * 									refresh: (optional) to refresh the content
	 * 								the module should these properties:
	 * 									i18.menu_caption: caption of the menu entry
	 * 									pageGroup: (optional) if pages of the same cathegory should go in on submenu
	 * 									i18.group_caption: (optional) caption of the submenu
	 */
	this.addPage = function( module ) {
		
		var $menu = $('#' + menuId);
		
		var caption = module.i18.group_caption || module.i18.menu_caption;
		
		if (module.pageGroup) {
			
			var $item = $menu.find( 'LI[group="{0}"]'.format(module.pageGroup) );
			if ($item.length) {
				
				var $submenu = $menu.find( 'ul[group="{0}"]'.format(module.pageGroup) );
				if (!$submenu.length) {

					// use menu_caption in submenu
					caption = module.i18.menu_caption;
				
					// no submenu, but an item with this group already exists
					// => create a submenu and insert existing item
					$submenu = $(
							('<li class="has-dropdown"><a>{caption}</a><ul group="{group}" /><li>')
							.format({
								group:		module.pageGroup,
								caption:	module.i18.group_caption,
							})
					).insertAfter($item)
					.click(function(e) {
						e.stopPropagation();
						if ( Modernizr.touch && !$(this).hasClass('disabled') ) {
							$(this).toggleClass('open-dropdown');
						}
					});	
					
					// switch $item from group caption to menu_caption 
					$item.find('A').text($item.prop('module').i18.menu_caption);
					
					// move $item into submenu
					$menu = $submenu.find('UL').append($item);
					
				} else {

					// submenu exists
					$menu = $submenu;

				}
			
			}
		}

		$li = $(
				('<li id="{itemid}" {group}><a href="#">{caption}</a></li>')
				.format({
					itemid:		module.name,
					caption:	caption,
					group:		(module.pageGroup) ? 'group="{0}"'.format(module.pageGroup) : '',
				})
		)
		.prop('module', module)
		.appendTo($menu)
		.click( function(e) {
			
			e.stopPropagation();

			if ( !$(this).hasClass('disabled') ) {
				
				// close dropdown menu (end hover)
				if ( !Modernizr.touch ) {
					$groups = $(this).parents('UL[group]').addClass('close-dropdown');
					window.setTimeout(function() {
						$groups.removeClass('close-dropdown');
					}, 100);
				}
				
				if( currentModule != module ) {

					// switch to new module

					// first cleanup prev. active module
					self.cleanupActiveModule();

					// switch module
					currentModule = module;

					// set selected menu item
					$(this).addClass('sel').parents('LI').addClass('sel').removeClass('open-dropdown');

					// display modules content
					if (currentModule.show) {
						// call show with modules scope
						currentModule.show.call(currentModule);
					}

				} else {

					// refresh modules content
					if (currentModule.refresh) {
						// call refresh with modules scope
						currentModule.refresh.call(currentModule);
					}

				}

			}

		} );

		if (module.refresh) {
			$li.find('A').addClass('refreshable');
		}

	};

	/**
	 * adds a menu item
	 * called from modules, to add their pages
	 *
	 * @param {tmaModule}	module
	 * @param {String}		id				id of the menu item
	 * @param {String}		caption			caption of the menu item
	 * @param {Function}	clickHandler	method, that handles the click event
	 * @returns {jQueryObject}
	 */
	this.addItem = function(module, id, caption, clickHandler) {

		return $('<li id="' + id + '"><a>' + (caption || '&nbsp;') + '</a></li>')
		.appendTo($('#' + menuId))
		.click( function(e) {

			if ( !$(this).hasClass('disabled') ) {

				clickHandler.call(module, e);

			}

		} );

	};

	/**
	 * removes a menu item
	 * called from modules, to cleanup their config item
	 *
	 * @param {String}		itemId			id of the menu item to remove
	 */
	this.removeItem = function( itemId ) {

		$('#' + menuId + ' LI#' + itemId).remove();

	};

	/**
	 * adds a menu item after another Item
	 * called from modules, to add their config item
	 *
	 * @param {String}		afterId			id of the menu item, after which the new item should be inserted
	 * @param {tmaModule}	module
	 * @param {String}		id				id of the menu item
	 * @param {String}		caption			caption of the menu item
	 * @param {Function}	clickHandler	method, that handles the click event
	 * @returns {jQueryObject}
	 */
	this.insertItemAfter = function(afterId, module, id, caption, clickHandler) {

		return $('<li id="' + id + '"><a>' + (caption || '&nbsp;') + '</a></li>')
		.insertAfter($('#' + menuId + ' LI#' + afterId))
		.click( function(e) {

			if ( !$(this).hasClass('disabled') ) {

				clickHandler.call(module, e);

			}

		} );

	};

	/**
	 * call the cleanup method of the prev. active module
	 */
	this.cleanupActiveModule = function() {

		if ( (currentModule) && (currentModule.cleanup) ) {
			// call cleanup with modules scope
			currentModule.cleanup.call(currentModule);
		}

		// reset record counter to 0
		$("#counter").text( tma.i18.record_counter.format( 0 ) );

		// clear config
		$('#config').empty().hide();

		// clear content
		$('#content').empty();

		// reset selected menu item
		$('#' + menuId + ' LI.sel').removeClass('sel');

	};

	this.enableItem = function(itemId) {
		$('#' + menuId + ' LI#' + itemId).removeClass('disabled');
	};

	this.disableItem = function(itemId) {
		$('#' + menuId + ' LI#' + itemId).addClass('disabled');
	};

	this.showItem = function(itemId) {
		$('#' + menuId + ' LI#' + itemId).show();
	};

	this.hideItem = function(itemId) {
		$('#' + menuId + ' LI#' + itemId).hide();
	};

}
